package ses

import (
	"bytes"
	"context"
	"fmt"
	"github.com/aws/aws-sdk-go-v2/aws"
	"github.com/aws/aws-sdk-go-v2/config"
	"github.com/aws/aws-sdk-go-v2/service/ses"
	sesTypes "github.com/aws/aws-sdk-go-v2/service/ses/types"
	"github.com/prozz/aws-embedded-metrics-golang/emf"
	"gitlab.bob.co.za/bob-public-utils/bobgroup-go-utils/errors"
	"gitlab.bob.co.za/bob-public-utils/bobgroup-go-utils/logs"
	"gitlab.bob.co.za/bob-public-utils/bobgroup-go-utils/s3"
	"gitlab.bob.co.za/bob-public-utils/bobgroup-go-utils/utils"
	"gopkg.in/gomail.v2"
	"io"
	"strings"
)

const (
	InvalidParameterValueErrorString = "InvalidParameterValue"
)

var (
	clients = map[string]*ClientWithHelpers{}

	ErrorSESClientNotEstablished = errors.Error("could not establish an SES client")
)

type ClientWithHelpers struct {
	SESClient *ses.Client
}

type Email struct {
	Recipient           string                 `json:"recipient"`
	FromName            string                 `json:"from_name"`
	FromEmail           string                 `json:"from_email"`
	Subject             string                 `json:"subject"`
	EmailCcName         string                 `json:"email_cc_name"`
	EmailCcAddress      string                 `json:"email_cc_address"`
	ReplyToEmailAddress string                 `json:"reply_to_email_address"`
	Body                EmailBody              `json:"body"`
	Attachments         *[]s3.S3UploadResponse `json:"attachments"`
}

type EmailBody struct {
	HTMLBody string `json:"html_body"`
	TextBody string `json:"text_body"`
}

func GetClient(region ...string) *ClientWithHelpers {
	sesRegion := "eu-west-1"

	// Set custom region
	if region != nil && len(region) > 0 {
		sesRegion = region[0]
	}

	// Check if client exists for region, if it does return it
	if sesClient, ok := clients[sesRegion]; ok && sesClient != nil {
		return sesClient
	}

	cfg, err := config.LoadDefaultConfig(context.TODO(),
		config.WithRegion(sesRegion),
	)
	if err != nil {
		return &ClientWithHelpers{}
	}

	sesClient := NewClient(cfg)

	clients[sesRegion] = sesClient
	return sesClient
}

func NewClient(cfg aws.Config) *ClientWithHelpers {
	return &ClientWithHelpers{
		SESClient: ses.NewFromConfig(cfg),
	}
}

func (c *ClientWithHelpers) SendEmail(email Email, emfNamespace string, isDebug bool) (*ses.SendRawEmailOutput, error) {
	if c.SESClient == nil {
		return nil, ErrorSESClientNotEstablished
	}

	recipients := utils.SplitAndCleanEmailAddresses(email.Recipient)
	if len(recipients) == 0 {
		return nil, nil
	}

	// Create raw message
	msg := gomail.NewMessage()
	msg.SetHeader("To", recipients...)

	// Set the from header, and use gomail's internal function if we have a name and email
	from := email.FromEmail
	if email.FromName != "" {
		from = msg.FormatAddress(email.FromEmail, email.FromName)
	}
	msg.SetHeader("From", from)
	msg.SetAddressHeader("Cc", email.EmailCcAddress, email.EmailCcName)
	msg.SetHeader("Subject", email.Subject)
	msg.SetBody("text/plain", email.Body.TextBody) // See comment for `AddAlternative`
	msg.AddAlternative("text/html", email.Body.HTMLBody)

	// Set the reply-to address
	if email.ReplyToEmailAddress != "" {
		msg.SetHeader("Reply-To", email.ReplyToEmailAddress)
	}

	// Add attachments
	if email.Attachments != nil && len(*email.Attachments) != 0 {
		for _, attachment := range *email.Attachments {
			err := addAttachmentsToEmail(msg, attachment, isDebug)
			if err != nil {
				return nil, err
			}
		}
	}

	// create a new buffer to add raw data
	var emailRaw bytes.Buffer
	_, err := msg.WriteTo(&emailRaw)
	if err != nil {
		return nil, err
	}

	// Append the CC address to the recipients
	if email.EmailCcAddress != "" {
		recipients = append(recipients, email.EmailCcAddress)
	}

	// create new raw message
	rawEmailInput := &ses.SendRawEmailInput{
		Source:       &from,
		Destinations: recipients,
		RawMessage: &sesTypes.RawMessage{
			Data: emailRaw.Bytes(),
		},
	}

	// Send it
	result, err := c.SESClient.SendRawEmail(context.TODO(), rawEmailInput)
	if err != nil {
		messageRejectedErrorCode := errors.AWSErrorExceptionCode(&sesTypes.MessageRejected{})
		mailFromDomainNotVerifiedErrorCode := errors.AWSErrorExceptionCode(&sesTypes.MailFromDomainNotVerifiedException{})
		configurationSetDoesNotExistErrorCode := errors.AWSErrorExceptionCode(&sesTypes.ConfigurationSetDoesNotExistException{})

		switch errors.AWSErrorExceptionCode(err) {
		case messageRejectedErrorCode, mailFromDomainNotVerifiedErrorCode, configurationSetDoesNotExistErrorCode:
			logs.ErrorWithMsg(errors.AWSErrorExceptionCode(err), err)
		case InvalidParameterValueErrorString:
			// Ignore errors due to invalid email addresses - we don't want to log it to raygun
			logs.Warn(err.Error())
			err = nil
		default:
			logs.ErrorWithMsg("Failed to send email", err)
		}

		if emfNamespace != "" {
			emf.New().
				Namespace(emfNamespace).
				MetricAs("email_failed", 1, emf.Count).
				Log()
		}
	} else {
		if emfNamespace != "" {
			emf.New().
				Namespace(emfNamespace).
				MetricAs("email_sent", 1, emf.Count).
				Log()
		}
	}

	return result, err
}

func addAttachmentsToEmail(msg *gomail.Message, attachment s3.S3UploadResponse, isDebug bool) error {
	if attachment.Bucket == "" || attachment.Filename == "" {
		logs.Warn("empty bucket and/or filename, cannot attach file")
		return nil
	}

	// Get object from S3
	object, err := s3.GetClient(isDebug).GetObject(attachment.Bucket, attachment.Filename, isDebug)
	if err != nil {
		logs.ErrorWithMsg(fmt.Sprintf("Failed to get S3 object from bucket %s filename %s", attachment.Bucket, attachment.Filename), err)
		return err
	}
	defer object.Body.Close()

	attachmentBytes, err := io.ReadAll(object.Body)
	if err != nil {
		logs.ErrorWithMsg("Could not encode attachment for email", err)
		return err
	}

	// Get the filename
	splitString := strings.Split(attachment.Filename, "/")
	filename := splitString[len(splitString)-1]

	msg.Attach(
		filename,
		gomail.SetCopyFunc(func(w io.Writer) error {
			_, err := w.Write(attachmentBytes)
			return err
		}),
		gomail.SetHeader(map[string][]string{"Content-Type": {*object.ContentType}}),
	)

	return nil
}
